from typing import Union, Type

import torch
import torch.nn as nn
import torch.nn.functional as F

from jbag.config import Config
from jbag.models.network_weight_initialization import initialize_network
from jbag.models.utils import get_norm_op, get_non_linear_op


class ConvBlock(nn.Module):
    def __init__(self, input_channels, mid_channels, output_channels,
                 norm_op: Union[None, Type[nn.Module]] = None,
                 norm_op_kwargs: dict = None,
                 non_linear: Union[None, Type[nn.Module]] = None,
                 non_linear_kwargs: dict = None,
                 ):
        super().__init__()
        if norm_op_kwargs is None:
            norm_op_kwargs = {}
        if non_linear_kwargs is None:
            non_linear_kwargs = {}

        self.activation = non_linear(**non_linear_kwargs)
        self.conv1 = nn.Conv2d(input_channels, mid_channels, kernel_size=3, padding=1, bias=True)
        self.norm1 = norm_op(mid_channels, **norm_op_kwargs)
        self.conv2 = nn.Conv2d(mid_channels, output_channels, kernel_size=3, padding=1, bias=True)
        self.norm2 = norm_op(output_channels, **norm_op_kwargs)

    def forward(self, x):
        x = self.conv1(x)
        x = self.norm1(x)
        x = self.activation(x)

        x = self.conv2(x)
        x = self.norm2(x)
        output = self.activation(x)

        return output


class UNetPlusPlus(nn.Module):
    def __init__(self, input_channels=1, num_classes=1,
                 norm_op: Union[None, Type[nn.Module]] = None,
                 norm_op_kwargs: dict = None,
                 non_linear: Union[None, Type[nn.Module]] = None,
                 non_linear_kwargs: dict = None,
                 ):
        super(UNetPlusPlus, self).__init__()
        self.output_channels = num_classes

        n1 = 64
        filters = [n1, n1 * 2, n1 * 4, n1 * 8, n1 * 16]

        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.up = nn.Upsample(scale_factor=2, mode="bilinear", align_corners=True)

        self.conv0_0 = ConvBlock(input_channels, filters[0], filters[0],
                                 norm_op=norm_op, norm_op_kwargs=norm_op_kwargs,
                                 non_linear=non_linear, non_linear_kwargs=non_linear_kwargs)
        self.conv1_0 = ConvBlock(filters[0], filters[1], filters[1],
                                 norm_op=norm_op, norm_op_kwargs=norm_op_kwargs,
                                 non_linear=non_linear, non_linear_kwargs=non_linear_kwargs)
        self.conv2_0 = ConvBlock(filters[1], filters[2], filters[2],
                                 norm_op=norm_op, norm_op_kwargs=norm_op_kwargs,
                                 non_linear=non_linear, non_linear_kwargs=non_linear_kwargs)
        self.conv3_0 = ConvBlock(filters[2], filters[3], filters[3],
                                 norm_op=norm_op, norm_op_kwargs=norm_op_kwargs,
                                 non_linear=non_linear, non_linear_kwargs=non_linear_kwargs)
        self.conv4_0 = ConvBlock(filters[3], filters[4], filters[4],
                                 norm_op=norm_op, norm_op_kwargs=norm_op_kwargs,
                                 non_linear=non_linear, non_linear_kwargs=non_linear_kwargs)

        self.conv0_1 = ConvBlock(filters[0] + filters[1], filters[0], filters[0],
                                 norm_op=norm_op, norm_op_kwargs=norm_op_kwargs,
                                 non_linear=non_linear, non_linear_kwargs=non_linear_kwargs)
        self.conv1_1 = ConvBlock(filters[1] + filters[2], filters[1], filters[1],
                                 norm_op=norm_op, norm_op_kwargs=norm_op_kwargs,
                                 non_linear=non_linear, non_linear_kwargs=non_linear_kwargs)
        self.conv2_1 = ConvBlock(filters[2] + filters[3], filters[2], filters[2],
                                 norm_op=norm_op, norm_op_kwargs=norm_op_kwargs,
                                 non_linear=non_linear, non_linear_kwargs=non_linear_kwargs)
        self.conv3_1 = ConvBlock(filters[3] + filters[4], filters[3], filters[3],
                                 norm_op=norm_op, norm_op_kwargs=norm_op_kwargs,
                                 non_linear=non_linear, non_linear_kwargs=non_linear_kwargs)

        self.conv0_2 = ConvBlock(filters[0] * 2 + filters[1], filters[0], filters[0],
                                 norm_op=norm_op, norm_op_kwargs=norm_op_kwargs,
                                 non_linear=non_linear, non_linear_kwargs=non_linear_kwargs)
        self.conv1_2 = ConvBlock(filters[1] * 2 + filters[2], filters[1], filters[1],
                                 norm_op=norm_op, norm_op_kwargs=norm_op_kwargs,
                                 non_linear=non_linear, non_linear_kwargs=non_linear_kwargs)
        self.conv2_2 = ConvBlock(filters[2] * 2 + filters[3], filters[2], filters[2],
                                 norm_op=norm_op, norm_op_kwargs=norm_op_kwargs,
                                 non_linear=non_linear, non_linear_kwargs=non_linear_kwargs)

        self.conv0_3 = ConvBlock(filters[0] * 3 + filters[1], filters[0], filters[0],
                                 norm_op=norm_op, norm_op_kwargs=norm_op_kwargs,
                                 non_linear=non_linear, non_linear_kwargs=non_linear_kwargs)
        self.conv1_3 = ConvBlock(filters[1] * 3 + filters[2], filters[1], filters[1],
                                 norm_op=norm_op, norm_op_kwargs=norm_op_kwargs,
                                 non_linear=non_linear, non_linear_kwargs=non_linear_kwargs)

        self.conv0_4 = ConvBlock(filters[0] * 4 + filters[1], filters[0], filters[0],
                                 norm_op=norm_op, norm_op_kwargs=norm_op_kwargs,
                                 non_linear=non_linear, non_linear_kwargs=non_linear_kwargs)

        self.final = nn.Conv2d(filters[0], num_classes, kernel_size=1)

    def forward(self, x):
        x0_0 = self.conv0_0(x)
        x1_0 = self.conv1_0(self.pool(x0_0))
        x0_1 = self.conv0_1(torch.cat([x0_0, self.up(x1_0)], 1))

        x2_0 = self.conv2_0(self.pool(x1_0))
        x1_1 = self.conv1_1(torch.cat([x1_0, self.up(x2_0)], 1))
        x0_2 = self.conv0_2(torch.cat([x0_0, x0_1, self.up(x1_1)], 1))

        x3_0 = self.conv3_0(self.pool(x2_0))

        xup_3_0 = self.up(x3_0)
        if x2_0.size()[2:] != xup_3_0.size()[2:]:
            xup_3_0 = F.interpolate(xup_3_0,  x2_0.size()[2:], mode='nearest')

        x2_1 = self.conv2_1(torch.cat([x2_0, xup_3_0], 1))
        x1_2 = self.conv1_2(torch.cat([x1_0, x1_1, self.up(x2_1)], 1))
        x0_3 = self.conv0_3(torch.cat([x0_0, x0_1, x0_2, self.up(x1_2)], 1))

        x4_0 = self.conv4_0(self.pool(x3_0))

        xup_4_0 = self.up(x4_0)
        if x3_0.size()[2:] != xup_4_0.size()[2:]:
            xup_4_0 = F.interpolate(xup_4_0, x3_0.size()[2:], mode='nearest')
        x3_1 = self.conv3_1(torch.cat([x3_0, xup_4_0], 1))

        xup_3_1 = self.up(x3_1)
        if x2_0.size()[2:] != xup_3_1.size()[2:]:
            xup_3_1 = F.interpolate(xup_3_1, x2_0.size()[2:], mode='nearest')
        x2_2 = self.conv2_2(torch.cat([x2_0, x2_1, xup_3_1], 1))

        x1_3 = self.conv1_3(torch.cat([x1_0, x1_1, x1_2, self.up(x2_2)], 1))
        x0_4 = self.conv0_4(torch.cat([x0_0, x0_1, x0_2, x0_3, self.up(x1_3)], 1))

        output = self.final(x0_4)
        return output


def build_unet_plus_plus(network_config: Config):
    norm_op = get_norm_op(network_config.norm_op, network_config.conv_dim)
    non_linear_op = get_non_linear_op(network_config.non_linear)

    params = {"input_channels": network_config.input_channels,
              "num_classes": network_config.num_classes,
              "norm_op": norm_op,
              "norm_op_kwargs": network_config.norm_op_kwargs,
              "non_linear": non_linear_op,
              "non_linear_kwargs": network_config.non_linear_kwargs,
              }
    network = UNetPlusPlus(**params)

    initialize_network(network, network_config)
    return network



import numpy as np
from skimage.measure import label, regionprops


def get_maximum_connectivity_region(data):
    labeled_image = label(data, connectivity=2)
    regions = regionprops(labeled_image)
    if regions:
        largest_region = max(regions, key=lambda r: r.area)
        max_connectivity_region = np.zeros_like(data, dtype=np.uint8)
        indices = largest_region.coords
        indices = indices.transpose()
        max_connectivity_region[*indices] = 1
        return max_connectivity_region
    return np.zeros_like(data, dtype=np.uint8)



import multiprocessing as mp
from concurrent.futures import ProcessPoolExecutor, as_completed
from typing import Union

from tqdm import tqdm

from jbag import logger


def execute(fn,
            processes: int,
            starargs: Union[list[tuple], tuple[tuple]] = (),
            starkwargs: Union[list[dict], tuple[dict]] = (),
            mp_context_method=None,
            show_progress_bar=True):
    """
    Run parallel processing using concurrent ProcessPoolExecutor.
    Args:
        fn (function): Function to be run.
        processes (int): Number of parallel processes.
        starargs (sequence, optional, default=()): Groups of args to be passed to `fn`.
        starkwargs (sequence, optional, default=()): Groups of kwargs to be passed to `fn`.
        mp_context_method (string, optional, default=None): Method of multiprocessing context.
            If None, use default mp context of operation system.
        show_progress_bar (bool, optional, default=True): Whether to show progress bar.

    Returns:

    """
    if not mp_context_method in ["fork", "spawn", "forkserver", None]:
        raise ValueError(
            f"Unsupported multi-processing context method: {mp_context_method}. Supported methods: fork, spawn, forkserver, and None for default")

    n_starargs = len(starargs)
    n_starkwds = len(starkwargs)
    assert (n_starargs > 0 and n_starkwds == 0) or (n_starargs == 0 and n_starkwds > 0) or (n_starargs == n_starkwds), \
        "The group number of parameters of args and kwargs don't match."

    if n_starargs == 0 and n_starkwds != 0:
        starargs = [()] * n_starkwds

    if n_starargs != 0 and n_starkwds == 0:
        starkwargs = [{}] * n_starargs

    if processes > mp.cpu_count():
        logger.warn(f"Requested number of processes {processes} is greater than the total number of CPU processes. "
                    f"Set the number of processes to the number of CPU processes {mp.cpu_count()}.")
        processes = mp.cpu_count()

    if processes > len(starargs):
        logger.warn(f"Request number of processes {processes} is greater than the number of tasks {len(starargs)}. "
                    f"Set number of processes to {len(starargs)}.")
        processes = len(starargs)

    features = []
    r = []
    mp_context = mp.get_context(mp_context_method) if mp_context_method is not None else None
    with ProcessPoolExecutor(max_workers=processes, mp_context=mp_context) as executor:
        for args, kwargs in zip(starargs, starkwargs):
            features.append(executor.submit(fn, *args, **kwargs))

        with tqdm(total=len(starargs), disable=not show_progress_bar) as pbar:
            for future in as_completed(features):
                r.append(future.result())
                pbar.update()
    return r



from collections import OrderedDict

import numpy as np
import torch


class PreloadedDataLoader:
    def __init__(self, dataset, batch_size):
        self.batch_size = batch_size
        self.dataset = iter(dataset)

    def __next__(self):
        return self.load_batch()

    def __iter__(self):
        return self

    def load_batch(self):
        batch_dict = OrderedDict()
        batch_is_dict = False
        for _ in range(self.batch_size):
            batch = next(self.dataset)
            if isinstance(batch, np.ndarray):
                batch = torch.from_numpy(batch)
                if 0 not in batch_dict:
                    batch_dict[0] = []
                    batch_dict[0].append(batch)
            elif type(batch) in (int, tuple, set, list):
                for i, ele in enumerate(batch):
                    if isinstance(ele, np.ndarray):
                        ele = torch.from_numpy(ele)
                    if i not in batch_dict:
                        batch_dict[i] = []
                    batch_dict[i].append(ele)

            elif isinstance(batch, dict):
                batch_is_dict = True
                for k, v in batch.items():
                    if isinstance(v, np.ndarray):
                        v = torch.from_numpy(v)

                    if k not in batch_dict:
                        batch_dict[k] = []
                    batch_dict[k].append(v)

        if batch_is_dict:
            for k, v in batch_dict.items():
                if isinstance(v[0], torch.Tensor):
                    batch_dict[k] = torch.stack(v)
                else:
                    batch_dict[k] = v
            return batch_dict

        batch = []
        for v in batch_dict.values():
            if isinstance(v[0], torch.Tensor):
                batch.append(torch.stack(v))
            else:
                batch.append(v)
        return tuple(batch)



import os

from jbag.io import ensure_output_file_dir_existence
import dicom2nifti as d2n


def dicom2nifti(input_dicom_series, output_nifti_file, pydicom_read_force=False):
    if not os.path.exists(input_dicom_series):
        raise ValueError(f"Input DICOM series {input_dicom_series} does not exist.")

    if pydicom_read_force:
        d2n.settings.pydicom_read_force = pydicom_read_force

    ensure_output_file_dir_existence(output_nifti_file)
    d2n.dicom_series_to_nifti(input_dicom_series, output_nifti_file, reorient_nifti=False)



from typing import Union

import torch

from jbag.transforms._utils import get_non_one_scalar
from jbag.transforms.transform import RandomTransform


class GammaTransform(RandomTransform):
    def __init__(self, keys,
                 apply_probability,
                 gamma: Union[tuple[float], list[float]],
                 p_invert_image: float,
                 synchronize_channels: bool = False,
                 p_per_channel: float = 1,
                 p_retain_stats: float = 1, ):
        """
        Gamma transform.
        Args:
            keys (str or sequence):
            apply_probability (float):
            gamma (sequence): Range for gamma value.
            p_invert_image(float): Probability of inverting output image.
            synchronize_channels (bool, optional, default=False): If True, use the same parameters for all channels.
            p_per_channel (float, optional, default=1): Probability of applying transform to each channel.
            p_retain_stats (float, optional, default=1): Probability of retaining data statistics.
        """
        assert len(gamma) == 2 and gamma[1] >= gamma[0]
        super().__init__(keys, apply_probability)
        self.gamma = gamma
        self.p_invert_image = p_invert_image
        self.synchronize_channels = synchronize_channels
        self.p_per_channel = p_per_channel
        self.p_retain_stats = p_retain_stats

    def _call_fun(self, data):
        apply_to_channel = torch.where(torch.rand(len(self.keys)) < self.p_per_channel)[0]
        if len(apply_to_channel) == 0:
            return data
        retain_stats = torch.rand(len(apply_to_channel)) < self.p_retain_stats
        invert_images = torch.rand(len(apply_to_channel)) < self.p_invert_image
        if self.synchronize_channels:
            gamma = [get_non_one_scalar(self.gamma), ] * len(apply_to_channel)
        else:
            gamma = [get_non_one_scalar(self.gamma) for _ in range(len(apply_to_channel))]

        for c, r, i, g in zip(apply_to_channel, retain_stats, invert_images, gamma):
            value = data[self.keys[c]]
            if i:
                value *= -1
            if r:
                mean_intensity = value.mean()
                std_intensity = value.std()
            min_intensity = value.min()
            intensity_range = value.max() - min_intensity
            value = ((value - min_intensity) / intensity_range.clamp(min=1e-7)).pow(g) * intensity_range + min_intensity
            if r:
                mean_here = value.mean()
                std_here = value.std()
                value -= mean_here
                value *= (std_intensity / std_here.clamp(min=1e-7))
                value += mean_intensity
            if i:
                value *= -1
            data[self.keys[c]] = value
        return data



# Largest Image Pixel Value
Largest_Image_Pixel_Value = (0x0028, 0x0107)
# Smallest Image Pixel Value
Smallest_Image_Pixel_Value = (0x0028, 0x0106)
# Rescale Slope
Rescale_Slope = (0x0028, 0x1053)
# Rescale Intercept
Rescale_Intercept = (0x0028, 0x1052)
# Pixel Padding Value
Pixel_Padding_Value = (0x0028,0x0120)



import numpy as np

from jbag.transforms.transform import Transform


class ZscoreNormalization(Transform):
    def __init__(self, keys, mean, std, lower_bound=None, upper_bound=None):
        """
        Z-score normalization.

        Args:
            keys (str or sequence):
            mean (float):
            std (float):
            lower_bound (float, optional, default=None):
            upper_bound (float, optional, default=None):
        """

        super().__init__(keys)
        self.mean = mean
        self.std = std
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def _call_fun(self, data):
        for key in self.keys:
            value = data[key]
            if self.lower_bound is not None or self.upper_bound is not None:
                np.clip(value, self.lower_bound, self.upper_bound, out=value)
            value -= self.mean
            value /= max(self.std, 1e-8)
            data[key] = value
        return data


class MinMaxNormalization(Transform):
    def __init__(self, keys, lower_bound_percentile=1, upper_bound_percentile=99):
        """
        Perform min-max normalization.

        Args:
            keys (str or sequence):
            lower_bound_percentile (int, optional, default=1):
            upper_bound_percentile (int, optional, default=99):
        """
        super().__init__(keys)
        self.lower_bound_percentile = lower_bound_percentile
        self.upper_bound_percentile = upper_bound_percentile

    def _call_fun(self, data):
        for key in self.keys:
            image = data[key]
            min_value, max_value = np.percentile(image, (self.lower_bound_percentile, self.upper_bound_percentile))
            image = (image - min_value) / (max_value - min_value)
            data[key] = image
        return data



from typing import Union

import numpy as np
import torch
from scipy.ndimage import fourier_gaussian
from torch.nn.functional import grid_sample

from jbag.samplers.utils import central_crop
from jbag.transforms._utils import get_scalar
from jbag.transforms.transform import RandomTransform


class SpatialTransform(RandomTransform):
    def __init__(self, keys,
                 apply_probability,
                 patch_size=None,
                 patch_center_dist_from_border: Union[int, list[int], tuple[int]] = 0,
                 random_crop: bool = False,
                 interpolation_modes: Union[str, list[str], tuple[str]] = "nearest",
                 p_rotation: float = 0.0,
                 rotation_angle_range: Union[float, list[float], tuple[float, float]] = (0, 2 * np.pi),
                 p_scaling: float = 0.0,
                 scaling_range: Union[float, list[float], tuple[float, float]] = (0.7, 1.3),
                 p_synchronize_scaling_across_axes: float = 1,
                 p_elastic_deform: float = 0.0,
                 elastic_deform_scale: Union[float, list[float], tuple[float, float]] = (0, 0.2),
                 elastic_deform_magnitude: Union[float, list[float], tuple[float, float]] = (0, 0.2),
                 ):
        """
        Spatial affine transform. The input data should be in the dimension format of [C, [H, W, D]].
        Args:
            keys (str or sequence):
            apply_probability (float):
            patch_size (sequence or None, optional, default=None):
            random_crop (bool, optional, default=False): If True, crop patch image randomly.
            interpolation_modes (str or sequence(str), optional, default="nearest"): interpolation modes for value of keys.
            Supported options "nearest" | "bilinear".
            p_rotation (float, optional, default=0): Probability of applying rotation.
            rotation_angle_range (float or sequence, optional, default=[0, 2 * np.pi]):
            p_scaling (float, optional, default=0): Probability of applying scaling.
            scaling_range (float or sequence, optional, default=[0.7, 1.3]):
            p_synchronize_scaling_across_axes (float, optional, default=1):
            p_elastic_deform (float, optional, default=0): Probability of applying elastic deform.
            elastic_deform_scale (float or sequence, optional, default=[0, 0.2]):
            elastic_deform_magnitude (float or sequence, optional, default=[0, 0.2]):

        """
        super().__init__(keys, apply_probability)
        assert patch_center_dist_from_border is not None

        if patch_size is not None:
            patch_size = tuple(patch_size)
        self.patch_size = patch_size
        if patch_size is not None and not isinstance(patch_center_dist_from_border, (tuple, list)):
            patch_center_dist_from_border = [patch_center_dist_from_border] * len(patch_size)
        self.patch_center_dist_from_border = patch_center_dist_from_border

        self.p_rotation = p_rotation
        self.p_scaling = p_scaling
        self.p_elastic_deform = p_elastic_deform
        self.random_crop = random_crop
        if isinstance(interpolation_modes, str):
            interpolation_modes = [interpolation_modes] * len(self.keys)

        self.interpolation_modes = interpolation_modes

        self.rotation_angle_range = rotation_angle_range
        self.scaling = scaling_range
        self.p_synchronize_scaling_across_axes = p_synchronize_scaling_across_axes
        self.elastic_deform_scale = elastic_deform_scale
        self.elastic_deform_magnitude = elastic_deform_magnitude

    def _call_fun(self, data):
        # the process here comes from nnUnet v2
        do_grid_sampling = False
        original_spatial_shape = data[self.keys[0]].shape[1:]
        spatial_dims = len(original_spatial_shape)
        patch_size = self.patch_size if self.patch_size is not None else original_spatial_shape
        if len(original_spatial_shape) != len(patch_size):
            raise ValueError(
                f"Patch size dimensions ({len(patch_size)}) must match original spatial dimensions ({len(original_spatial_shape)}) of data.")

        do_rotation = np.random.random_sample() < self.p_rotation
        do_scale = np.random.random_sample() < self.p_scaling
        do_deform = np.random.random_sample() < self.p_elastic_deform
        if do_rotation:
            angles = [get_scalar(self.rotation_angle_range) for _ in range(spatial_dims)]
        else:
            angles = [0] * spatial_dims

        if do_scale:
            scales = [get_scalar(self.scaling) for _ in
                      range(spatial_dims)] if np.random.random_sample() < self.p_synchronize_scaling_across_axes \
                else [get_scalar(self.scaling)] * spatial_dims
        else:
            scales = [1] * spatial_dims

        if do_scale or do_rotation:
            do_grid_sampling = True
            if spatial_dims == 3:
                affine = create_affine_matrix_3d(angles, scales)
            elif spatial_dims == 2:
                affine = create_affine_matrix_2d(angles[0], scales)
            else:
                raise RuntimeError(f"Unsupported dimension: {spatial_dims}")
        else:
            affine = None  # this will allow us to detect that we can skip computations

        if do_deform:
            do_grid_sampling = True
            grid_scale = [i / j for i, j in zip(original_spatial_shape, patch_size)]
            deformation_scales = [
                get_scalar(self.elastic_deform_scale) for _ in
                range(spatial_dims)] if np.random.random_sample() < self.p_synchronize_scaling_across_axes \
                else [get_scalar(self.elastic_deform_scale)] * spatial_dims
            # sigmas must be in pixels, as this will be applied to the deformation field
            sigmas = [i * j for i, j in zip(deformation_scales, patch_size)]
            # the magnitude of the deformation field must adhere to the torch"s value range for grid_sample, i.e. [-1. 1] and not pixel coordinates. Do not use sigmas here
            # we need to correct magnitude by grid_scale to account for the fact that the grid will be wrt to the image size but the magnitude should be wrt the patch size. oof.
            magnitude = [
                get_scalar(self.elastic_deform_magnitude) / grid_scale[i] for i in range(spatial_dims)]
            # doing it like this for better memory layout for blurring
            elastic_offsets = torch.normal(mean=0, std=1, size=(spatial_dims, *patch_size))

            # all the additional time elastic deform takes is spent here
            for d in range(spatial_dims):
                # np fft is faster than torch
                tmp = np.fft.fftn(elastic_offsets[d].numpy())
                tmp = fourier_gaussian(tmp, sigmas)
                elastic_offsets[d] = torch.from_numpy(np.fft.ifftn(tmp).real)

                mx = torch.max(torch.abs(elastic_offsets[d]))
                elastic_offsets[d] /= (mx / np.clip(magnitude[d], a_min=1e-8, a_max=np.inf))
            elastic_offsets = torch.permute(elastic_offsets, (1, 2, 3, 0))
        else:
            elastic_offsets = None

        if self.random_crop:
            for i in range(len(original_spatial_shape)):
                assert patch_size[i] + self.patch_center_dist_from_border[i] * 2 <= original_spatial_shape[i], \
                    f"patch_size in spatial dimension {i} is {patch_size[i]}, patch center distance from border is {self.patch_center_dist_from_border[i]}, original shape {original_spatial_shape[i]} does not match the size requirement."
            center_location = []
            for i in range(len(original_spatial_shape)):
                dist_from_boarder = self.patch_center_dist_from_border[i]
                mn = patch_size[i] // 2 + dist_from_boarder
                mx = original_spatial_shape[i] - patch_size[i] // 2 - self.patch_center_dist_from_border[i]
                assert mn <= mx
                center_location.append(np.random.randint(mn, mx))
        else:
            center_location = [i // 2 for i in original_spatial_shape]

        if do_grid_sampling:
            grid = _create_grid(patch_size)
            grid_scale = torch.Tensor([i / j for i, j in zip(original_spatial_shape, patch_size)])
            grid /= grid_scale

            if elastic_offsets is not None:
                grid += elastic_offsets
            if affine is not None:
                grid = torch.matmul(grid, torch.from_numpy(affine).float())

            mn = grid.mean(dim=list(range(len(original_spatial_shape))))
            new_center = torch.Tensor(
                [(j / (i / 2) - 1) for i, j in zip(original_spatial_shape, center_location)])
            grid += - mn + new_center

        for i, key in enumerate(self.keys):
            value = data[key]
            # check if affine transform is needed
            if do_grid_sampling:
                dtype = value.dtype
                if dtype is not torch.float32:
                    value = value.to(torch.float32)
                value = grid_sample(value[None], grid[None], mode=self.interpolation_modes[i], padding_mode="zeros",
                                    align_corners=False).to(dtype)[0]
            else:
                if patch_size != original_spatial_shape:
                    value = central_crop(value, center_location, self.patch_size)
            data[key] = value
            if do_grid_sampling:
                data["spatial"] = True
            else:
                data["spatial"] = False
        return data


def _create_grid(size: list[int]):
    space = [torch.linspace((-s + 1) / s, (s - 1) / s, s) for s in size[::-1]]
    grid = torch.meshgrid(space, indexing="ij")
    grid = torch.stack(grid, -1)
    spatial_dims = list(range(len(size)))
    grid = grid.permute((*spatial_dims[::-1], len(size)))
    return grid


def create_affine_matrix_3d(rotation_angles, scaling_factors):
    # Rotation matrices for each axis
    rotation_x = np.array([[1, 0, 0],
                           [0, np.cos(rotation_angles[0]), -np.sin(rotation_angles[0])],
                           [0, np.sin(rotation_angles[0]), np.cos(rotation_angles[0])]])

    rotation_y = np.array([[np.cos(rotation_angles[1]), 0, np.sin(rotation_angles[1])],
                           [0, 1, 0],
                           [-np.sin(rotation_angles[1]), 0, np.cos(rotation_angles[1])]])

    rotation_z = np.array([[np.cos(rotation_angles[2]), -np.sin(rotation_angles[2]), 0],
                           [np.sin(rotation_angles[2]), np.cos(rotation_angles[2]), 0],
                           [0, 0, 1]])

    # Scaling matrix
    scaling = np.diag(scaling_factors)

    # Combine rotation and scaling
    affine = rotation_z @ rotation_y @ rotation_x @ scaling
    return affine


def create_affine_matrix_2d(rotation_angle, scaling_factors):
    # Rotation matrix
    rotation = np.array([[np.cos(rotation_angle), -np.sin(rotation_angle)],
                         [np.sin(rotation_angle), np.cos(rotation_angle)]])

    # Scaling matrix
    scaling = np.diag(scaling_factors)

    # Combine rotation and scaling
    affine = rotation @ scaling
    return affine



import jbag.metrics.SDF
from jbag.transforms.transform import Transform


class SDF(Transform):
    def __init__(self, keys, normalize):
        """
        Compute signed distance function.

        Args:
            keys (str or sequence): Binary segmentation for computing SDF.
            normalize (bool): If `True`, normalize the SDF by min-max normalization.
        """
        super().__init__(keys)
        self.normalize = normalize

    def _call_fun(self, data):
        for key in self.keys:
            segmentation = data[key]
            sdf_map = jbag.metrics.SDF.SDF(segmentation, self.normalize)
            data[f"{key}_SDF"] = sdf_map
        return data



from typing import Callable

import numpy as np


class AgglomerativeHierarchicalClustering:
    def __init__(self, metric: Callable[[np.ndarray, np.ndarray], float], linkage: str = "complete",
                 n_clusters: int = 1):
        """
        Agglomerative hierarchical clustering algorithm. Only "complete" linkage is supported for now.
        Note that `sklearn.cluster.AgglomerativeClustering` is much faster, this implementation is used when
        AgglomerativeClustering in sklearn fails to meet the data dimension restriction when the sample is
        represented by a matrix/tensor with feature dimensions greater than 1. While AgglomerativeClustering in sklearn
        only supports data shape of (n_samples, n_features).

        Args:
            metric (Callable[[np.ndarray, np.ndarray], float]): metric function for measuring element distance.
            linkage (str, optional, default="complete"): Currently, only "complete" is supported.
            n_clusters  (int, optional, default=1): The number of clusters for clustering.
        """
        supported_linkages = ["complete"]
        if linkage not in supported_linkages:
            raise ValueError(f"linkage must be one of {supported_linkages}")
        if n_clusters < 1:
            raise ValueError(f"Minimum number of clusters must be >= 1")
        self.metric = metric
        self.linkage = linkage
        self.n_clusters = n_clusters

    def fit(self, X):
        """
        Perform agglomerative hierarchical clustering on `X`.
        Args:
            X (np.ndarray): The input data `X` is expected to have be (n_samples, ...).

        Returns:
        The clustering with desired number of clusters and the full grouping tree.
        """
        n_samples = X.shape[0]

        if n_samples < 2:
            raise ValueError(f"Need at least 2 samples for clustering.")

        cluster_distance = np.zeros((n_samples, n_samples))

        for i in range(n_samples):
            for j in range(i + 1, n_samples):
                distance = self.metric(X[i], X[j])
                cluster_distance[i][j] = distance
                cluster_distance[j][i] = distance

        np.fill_diagonal(cluster_distance, np.inf)

        k_steps = n_samples - self.n_clusters
        linkage_matrix = np.zeros((k_steps, 2), dtype=int)

        # active_cluster_indices is the unmerged cluster indices
        active_cluster_indices = list(range(n_samples))

        actual_cluster_ids = list(range(n_samples))
        new_cluster_id = n_samples

        for k_step in range(k_steps):

            argmin = np.argmin(cluster_distance)

            i_idx, j_idx = np.unravel_index(argmin, cluster_distance.shape)

            i_id, j_id = actual_cluster_ids[i_idx], actual_cluster_ids[j_idx]
            linkage_matrix[k_step][0] = i_id
            linkage_matrix[k_step][1] = j_id

            actual_cluster_ids[i_idx] = new_cluster_id
            new_cluster_id += 1

            active_cluster_indices.remove(j_idx)

            for i in active_cluster_indices:
                if i != i_idx:
                    distance = self._complete_linkage(cluster_distance, i_idx, j_idx, i)
                    cluster_distance[i][i_idx] = distance
                    cluster_distance[i_idx][i] = distance

            cluster_distance[j_idx, :] = np.inf
            cluster_distance[:, j_idx] = np.inf

        clusterings = self.get_clusters(linkage_matrix, n_samples)
        return clusterings[0], clusterings

    @staticmethod
    def get_clusters(linkage_matrix, n_samples: int):
        """
        Build the clusterings from linkage matrix. This function can also be applied to
        sklearn.cluster.AgglomerativeClustering for building the clusterings:

        Args:
            linkage_matrix:
            n_samples:

        Returns: Clusterings with different numbers of clusters.

        Examples:
        >>> from sklearn.cluster import AgglomerativeClustering
        >>> import numpy as np
        >>> X = np.array([[1, 2], [1, 4], [1, 0],[4, 2], [4, 4], [4, 0]])
        >>> clustering = AgglomerativeClustering(compute_full_tree=True).fit(X)
        >>> linkage_matrix = clustering.children_
        >>> clusterings = AgglomerativeHierarchicalClustering.get_clusters(linkage_matrix, X.shape[0])
        >>> clusterings
        [[[0, 1, 2, 3, 4, 5]], [[0, 1, 2], [3, 4, 5]], [[0, 1, 2], [3, 5], [4]], [[0, 1], [2], [3, 5], [4]], [[0, 1], [2], [3], [4], [5]], [[0], [1], [2], [3], [4], [5]]]
        """
        cluster_contents = [[i] for i in range(n_samples)]
        active_cluster_ids = list(range(n_samples))

        history_of_cluster_sets = []

        current_set_of_clusters = [[i] for i in range(n_samples)]
        history_of_cluster_sets.append(current_set_of_clusters)

        for i in range(linkage_matrix.shape[0]):
            c1_id, c2_id = linkage_matrix[i]

            # Retrieve the cluster members of merged clusters
            samples_c1 = cluster_contents[c1_id]
            samples_c2 = cluster_contents[c2_id]

            new_cluster_samples = sorted(samples_c1 + samples_c2)

            cluster_contents.append(new_cluster_samples)
            new_cluster_id = n_samples + i

            active_cluster_ids.remove(c1_id)
            active_cluster_ids.remove(c2_id)
            active_cluster_ids.append(new_cluster_id)

            current_set_snapshot = [list(cluster_contents[k]) for k in active_cluster_ids]
            current_set_snapshot.sort()
            history_of_cluster_sets.append(current_set_snapshot)

        return history_of_cluster_sets[::-1]

    @staticmethod
    def _complete_linkage(cluster_distance, linkage_cluster_i_idx, linkage_cluster_j_idx, cluster_idx):

        max_distance = max(cluster_distance[cluster_idx][linkage_cluster_i_idx],
                           cluster_distance[cluster_idx][linkage_cluster_j_idx])
        return max_distance



from typing import Union

import numpy as np
import torch


def get_margin(patch_size):
    """
    Determine the margin length that cannot be the central point of the patch of which cannot fit the size of patch_size.
    Return a len(patch_size) X 2 matrix with left and right (before and after, top and bottom) margin size.

    Args:
        patch_size (sequence):

    Returns:

    """
    if not isinstance(patch_size, np.ndarray):
        patch_size = np.asarray(patch_size)

    return np.where(patch_size % 2 == 1, np.asarray([patch_size // 2, patch_size // 2]),
                    np.asarray([patch_size / 2 - 1, patch_size / 2])).astype(np.int16).T


def get_non_margin_region(image_shape: tuple | list, patch_size: tuple | list):
    """
    Build a mask that cover the area which can be view as the central point of the patch with the size of
    patch_size. Mask the central area that can be extracted. Left the margin area.

    Args:
        image_shape (sequence):
        patch_size (sequence):

    Returns:

    """
    assert len(image_shape) == len(patch_size)
    margin = get_margin(patch_size)
    mask = np.zeros(image_shape, dtype=bool)
    mask[tuple([slice(j[0], i - j[1]) for i, j in zip(image_shape, margin)])] = 1
    return mask


def encode_one_hot(data):
    shape = data.shape
    data = data.flatten().astype(np.uint8)
    num_classes = int(data.max() + 1)

    result = np.zeros((data.size, num_classes), dtype=bool)
    index = np.arange(data.size)
    result[index, data] = 1
    result_shape = tuple(list(shape) + [num_classes])
    if result.shape != result_shape:
        result = result.reshape(result_shape)

    # move category axis to the first. [C, ...]
    result = np.moveaxis(result, -1, 0)
    return result


def central_crop(data: Union[np.ndarray, torch.Tensor], center, shape):
    """
    Crop patch with the shape from data at the central coordinate.

    Args:
        data (numpy.ndarray or torch.Tensor):
        center (sequence):
        shape (sequence):

    Returns:

    """
    assert len(data.shape) >= len(shape) == len(center)
    center = np.asarray(center)
    margin = get_margin(shape)[:, 0]
    # coordinate is the corner coordinate responding to the central.
    coordinate = center - margin
    assert (coordinate >= 0).all(), "center {} or shape {} error".format(center, shape)
    cropped = crop_patch(data, coordinate, shape)
    return cropped


def crop_patch(data: Union[np.ndarray, torch.Tensor], coordinate, patch_size):
    """
    Crop patch from `data` at the left/superior/anterior `coordinate`. If data has larger dimensions than shape
    dimensions N, the last N dimensions will be cropped.

    Args:
        data  (numpy.ndarray or torch.Tensor):
        coordinate (sequence):
        patch_size (sequence):

    Returns:

    """
    data_shape = data.shape
    assert len(data_shape) >= len(coordinate) == len(patch_size)
    end = coordinate + patch_size
    slice_sequence = [slice(i, j) for i, j in zip(coordinate, end)]
    # append the extra dimensions, batch/channel...
    slice_sequence = [slice(None)] * (len(data_shape) - len(end)) + slice_sequence
    cropped = data[tuple(slice_sequence)]
    return cropped



from abc import ABC, abstractmethod

import numpy as np


class Transform(ABC):
    def __init__(self, keys):
        if not keys:
            raise ValueError("Keys cannot be empty.")
        if isinstance(keys, str):
            keys = [keys]
        self.keys = keys

    def __call__(self, data):
        return self._call_fun(data)

    @abstractmethod
    def _call_fun(self, data):
        ...


class RandomTransform(Transform, ABC):
    def __init__(self, keys, apply_probability):
        super().__init__(keys)
        self.apply_probability = apply_probability

    def __call__(self, data):
        if np.random.random_sample() < self.apply_probability:
            return self._call_fun(data)
        else:
            return data



from typing import Union

from docx.oxml import OxmlElement
from docx.oxml.ns import qn
from docx.shared import Pt
from docx.table import _Cell, Table


def cm_to_dxa(cm):
    """
    Convert centimeters to twentieths of a point (dxa).
    """
    return int(cm * 1440 / 2.54)


def set_cell(cell: _Cell, text, font="Times New Roman", font_size=10, bold=False, italic=False,
             underline=False, cell_margins: list[float] = None):
    """
    Set cell text and properties.
    Args:
        cell (docx.table._Cell):
        text (str):
        font (str, optional, default="Times New Roman"):
        font_size (int, optional, default=10):
        bold (bool, optional, default=False):
        italic (bool, optional, default=False):
        underline (bool, optional, default=False):
        cell_margins (list, optional, default=[]): The cell margins should be in the format of [left, right, top, bottom] in the unit of centimeter.

    Returns:

    """
    cell.text = text
    run = cell.paragraphs[0].runs[0]
    run.font.name = font
    run.font.size = Pt(font_size)
    run.font.bold = bold
    run.font.italic = italic
    run.font.underline = underline
    if cell_margins:
        tc = cell._tc
        tcPr = tc.get_or_add_tcPr()

        tcMar = tcPr.find(qn("w:tcMar"))
        if tcMar is None:
            tcMar = OxmlElement("w:tcMar")
            tcPr.append(tcMar)

        margin_keys = ["left", "right", "top", "bottom"]
        for i, value in enumerate(cell_margins):
            if value is None:
                continue

            tag = "w:{}".format(margin_keys[i])
            margin_element = tcMar.find(qn(tag))
            if margin_element is None:
                margin_element = OxmlElement(tag)
                tcMar.append(margin_element)
            value = cm_to_dxa(value)
            margin_element.set(qn("w:w"), str(value))
            margin_element.set(qn("w:type"), "dxa")


def set_cell_border(cell: _Cell,
                    borders: Union[str, list[str], tuple[float, ...]],
                    styles: Union[str, list[str], tuple[float, ...]] = "single",
                    sizes: Union[float, list[float], tuple[float, ...]] = 4,
                    colors: Union[str, list[str], tuple[str, ...]] = "auto"):
    """

    Args:
        cell:
        borders: Option: top, bottom, left, right.
        styles:
        sizes: 4 is equal to 0.5 pt, 8 for 1 pt.

    Returns:

    """
    if isinstance(borders, str):
        borders = [borders]

    valid_borders = ["top", "bottom", "left", "right"]
    for border in borders:
        if border not in valid_borders:
            raise ValueError(f"Invalid border type: {border}. Supported are : {', '.join(valid_borders)}")

    if isinstance(styles, str):
        styles = [styles] * len(borders)
    if not isinstance(sizes, (list, tuple)):
        sizes = [sizes] * len(borders)
    if isinstance(colors, str):
        colors = [colors] * len(borders)

    assert len(borders) == len(styles) == len(sizes) == len(colors)

    tc = cell._tc
    tcPr = tc.get_or_add_tcPr()
    tcBorders = OxmlElement("w:tcBorders")

    for border, style, size, color in zip(borders, styles, sizes, colors):
        border_element = OxmlElement(f"w:{border}")
        border_element.set(qn("w:val"), style)
        border_element.set(qn("w:sz"), str(size))
        border_element.set(qn("w:color"), color)
        tcBorders.append(border_element)
    tcPr.append(tcBorders)


def set_three_line_border(table: Table,
                          outer_line_weight: Union[float, int] = 8,
                          inner_line_weight: Union[float, int] = 4):
    """
    Draw three line table borders.
    Args:
        table (docx.table.Table):
        outer_line_weight (float or int, optional, default=8): Line weight for top and bottom borders. Defaults are 1 pt.
        8 indicates 1 pt.
        inner_line_weight (float or int, optional, default=8): Inner line weight. Defaults are 0.5 pt. 4 indicates 0.5 pt.

    Returns:

    """
    first_row = table.rows[0]
    for cell in first_row.cells:
        set_cell_border(cell, borders=["top"], sizes=outer_line_weight)
        set_cell_border(cell, borders=["bottom"], sizes=inner_line_weight)

    last_row = table.rows[-1]
    for cell in last_row.cells:
        set_cell_border(cell, borders=["bottom"], sizes=outer_line_weight)



from scipy.ndimage import distance_transform_edt
from skimage.segmentation import find_boundaries


def SDF(X, normalize=False):
    """
    Compute signed distance function(SDF) of the input.

    Args:
        X (numpy.ndarray or torch.Tensor): Input data ndarray or tensor.
        normalize (bool, optional, default=False): If True, perform max-min normalization for SDF.
    """
    pos_distance = distance_transform_edt(X)
    neg_segmentation = ~X
    neg_distance = distance_transform_edt(neg_segmentation)

    boundary = find_boundaries(X, mode="inner")
    eps = 1e-6
    if normalize:
        sdf = (neg_distance - neg_distance.min()) / (neg_distance.max() - neg_distance.min() + eps) - \
              (pos_distance - pos_distance.min()) / (pos_distance.max() - pos_distance.min() + eps)
    else:
        sdf = neg_distance - pos_distance
    sdf[boundary] = 0

    return sdf



import os.path
from typing import Union

from docx import Document
from docx.shared import Inches
from docx.text.paragraph import Paragraph


def add_image(image_file,
              doc: Union[Document, None] = None,
              paragraph: Union[Paragraph, None] = None,
              width: float = 5,
              center: bool = True):
    if not os.path.isfile(image_file):
        raise FileNotFoundError(f"Input image file {image_file} does not exist.")

    assert not all([doc is None, paragraph is None])

    if paragraph is None:
        paragraph = doc.add_paragraph()

    run = paragraph.add_run()
    run.add_picture(image_file, width=Inches(width))
    if center:
        paragraph.alignment = 1



def combined_mean_std(mean1, std1, num1, mean2, std2, num2, ddof=1):
    """
    Calculate the mean and std of a group of data composed by two groups of data, given the mean, the std, and the
    number of data of the two subsets.

    Args:
          mean1 (float): Mean value of group1.
          std1 (float): Std value of group1.
          num1 (int): Number of data in group1.
          mean2 (float): Mean value of group2.
          std2 (float): Std value of group2.
          num2 (int): Number of data in group2.
          ddof (int, optional, default=1): Means of delta degrees of freedom. Default is 1 for unbiased estimation.
    """
    mean = (num1 * mean1 + num2 * mean2) / (num1 + num2)
    d = ((num1 - ddof) * std1 ** 2 + (num2 - ddof) * std2 ** 2) / (num1 + num2 - ddof) + \
        num1 * num2 * (mean1 - mean2) ** 2 / (num1 + num2) / (num1 + num2 - ddof)
    std = d ** 0.5
    return mean, std



from functools import partial

from torch import nn

from jbag import logger
from jbag.config import Config


def kaiming_initialize(module, a=1e-2, nonlinearity="leaky_relu"):
    """
    kaiming initialization. a is set to 0.01 by default, this is only worked for LeakyReLU.
    Args:
        module (torch.nn.Module):
        a (float, optional, default=1e-2): Set 0 for ReLu. Set 1e-2 for LeakyReLU.
        nonlinearity (str, optional, default="leaky_relu"):

    Returns:

    """
    if isinstance(module, (nn.Conv2d, nn.Conv3d, nn.ConvTranspose2d, nn.ConvTranspose3d)):
        module.weight = nn.init.kaiming_normal_(module.weight, a=a, nonlinearity=nonlinearity)
        if module.bias is not None:
            module.bias = nn.init.constant_(module.bias, 0)


def get_initialization_fn(method_name):
    match method_name:
        case "kaiming_initialize":
            return kaiming_initialize


def initialize_network(network: nn.Module, network_config: Config):
    allow_init = False
    network_initialization_params = None
    method = None
    if "initialization" in network_config and "allow_init" in network_config.initialization and "method" in network_config.initialization:
        allow_init = network_config.initialization.allow_init
        method = network_config.initialization.method

    if allow_init:
        if "params" in network_config.initialization:
            network_initialization_params = network_config.initialization.params.as_primitive()
        else:
            logger.warning("No initialization parameters were provided for network weight initialization.")

    if allow_init:
        logger.info(f"********Initializing network weights********")
        logger.info(f"Method {method}")
        logger.info(f"Params: {network_initialization_params}")
        logger.info(f"********Initializing network weights********")
        init_fn = partial(get_initialization_fn(method), **network_initialization_params)
        network.apply(init_fn)



import numpy as np


def get_scalar(scalar_type):
    if isinstance(scalar_type, (list, tuple)):
        assert scalar_type[1] >= scalar_type[0]
        if scalar_type[0] == scalar_type[1]:
            scalar = scalar_type[0]
        else:
            scalar = np.random.uniform(*scalar_type)
    else:
        scalar = scalar_type
    return scalar


def get_non_one_scalar(scalar_range):
    """
    Get a scalar uniformly sampled from [scalar_range[0], scalar_range[1]), but skip scalar of 1.
    Args:
        scalar_range:

    Returns:

    """
    assert scalar_range[1] >= scalar_range[0]
    if scalar_range[0] == scalar_range[1]:
        scalar = scalar_range[0]
    else:
        if scalar_range[1] <= 1:
            scalar = np.random.uniform(scalar_range[0], scalar_range[1])
        else:
            if np.random.random_sample() < 0.5 and scalar_range[0] < 1:
                scalar = np.random.uniform(scalar_range[0], 1)
            else:
                scalar = np.random.uniform(max(scalar_range[0], 1), scalar_range[1])
    return scalar


def get_max_spatial_dims(keys, valid_keys, data):
    max_spatial_dims = 0
    for c in valid_keys:
        value = data[keys[c]]
        spatial_dims = len(value.shape) - 1
        max_spatial_dims = max(max_spatial_dims, spatial_dims)
    return max_spatial_dims



from typing import Union

import torch

from jbag.transforms._utils import get_scalar
from jbag.transforms.transform import RandomTransform


class GaussianNoiseTransform(RandomTransform):
    def __init__(self, keys,
                 apply_probability,
                 noise_variance=Union[tuple[float], list[float]],
                 synchronize_channels: bool = False,
                 p_per_channel: float = 1):
        """
        Apply Gaussian noise to image.
        Args:
            keys (str or sequence):
            apply_probability (float):
            noise_variance (sequence): Range of noise variance.
            synchronize_channels (bool, optional, default=False): If True, use the same parameters for generating Gaussian noise.
            p_per_channel (float, optional, default=1): Probability of applying Gaussian noise for each channel.
        """
        super().__init__(keys, apply_probability)
        self.noise_variance = noise_variance
        self.synchronize_channels = synchronize_channels
        self.p_per_channel = p_per_channel

    def _call_fun(self, data):
        apply_to_channel = torch.where(torch.rand(len(self.keys)) < self.p_per_channel)[0]
        if len(apply_to_channel) == 0:
            return data
        if self.synchronize_channels:
            sigmas = [get_scalar(self.noise_variance), ] * len(apply_to_channel)
        else:
            sigmas = [get_scalar(self.noise_variance) for _ in range(len(apply_to_channel))]

        for c, sigma in zip(apply_to_channel, sigmas):
            value = data[self.keys[c]]
            image_shape = value.shape
            gaussian = torch.normal(0, sigma, image_shape)
            value += gaussian
            data[self.keys[c]] = value
        return data



from abc import ABC, abstractmethod

import numpy as np

from .utils import get_margin, encode_one_hot, get_non_margin_region


class CoordinateGenerator(ABC):

    @abstractmethod
    def __getitem__(self, index):
        ...

    @abstractmethod
    def __len__(self):
        ...

    @abstractmethod
    def regenerate(self):
        ...


class WeightedCoordinateGenerator(CoordinateGenerator):

    def __init__(self, num_coordinates, weights):
        """
        Choose coordinates according to the weight.

        Args:
            num_coordinates (int):
            weights (numpy.ndarray or torch.Tensor):
        """
        # number of coordinates that need to be picked.
        self.n = num_coordinates
        self.coordinate_system_shape = weights.shape
        self.a = weights.size

        p = weights / np.sum(weights)
        self.p = p.flatten()

        self.coordinates = self.get_coordinates()

    def __getitem__(self, index):
        return self.coordinates[index]

    def __len__(self):
        return self.n

    def get_coordinates(self):
        coordinates = np.random.choice(self.a, size=self.n, p=self.p)
        coordinates = np.asarray(np.unravel_index(coordinates, self.coordinate_system_shape))
        return coordinates.T

    def regenerate(self):
        self.coordinates = self.get_coordinates()


class BalancedCoordinateGenerator(WeightedCoordinateGenerator):
    def __init__(self, num_coordinates, label_map, patch_size):
        """
        Generate coordinates according to the proportion of categories. In general, the smaller the percentage of
        category, the more likely it is to be taken. Parameter data is the un-one-hot label.

        Args:
            num_coordinates (int):
            label_map (numpy.ndarray or torch.Tensor): The label matrix. this data will be needed to generate the weight map.
            patch_size (sequence):
        """
        self.n = num_coordinates
        self.patch_size = patch_size

        for i in range(len(label_map.shape)):
            assert label_map.shape[i] >= patch_size[
                i], f"Data size ({label_map.shape[i]}) along dimension {i} must not less than patch size ({patch_size[i]})."
        super().__init__(num_coordinates=num_coordinates, weights=self.get_weights(label_map))

    def get_weights(self, weight_label_map):
        # one-hot first
        label_onehot = encode_one_hot(weight_label_map)

        # exclude margin
        valid_region = get_non_margin_region(weight_label_map.shape, self.patch_size)
        weights = label_onehot * valid_region
        weights = weights.astype(np.float32)

        # normalize on every category axis
        category_sum = np.sum(weights, axis=tuple(range(1, len(weight_label_map.shape) + 1)), keepdims=True)
        weights = weights / (category_sum + 1)
        weights = np.sum(weights, axis=0).astype(np.float32)
        return weights


class GridCoordinateGenerator(CoordinateGenerator):
    def __init__(self, original_shape, patch_shape, valid_shape):
        """

        Args:
            original_shape (sequence): The shape of original image.
            patch_shape (sequence): The shape of one patch.
            valid_shape (sequence): The size of valid patch. Valid shape should be less than patch shape. A valid patch
                needs to be extracted from a patch.
        """
        self.original_shape = np.asarray(original_shape, dtype=np.uint32)
        self.patch_shape = np.asarray(patch_shape, dtype=np.uint32)
        self.valid_shape = np.asarray(valid_shape, dtype=np.uint32)

        self.padding_size = get_margin(patch_shape) - get_margin(valid_shape)
        self.valid_central_coordinates = self.get_coordinates()
        self.padded_coordinates = self.valid_central_coordinates + self.padding_size[:, 0][None, :]

    def get_coordinates(self):
        """
        Get central coordinates according to the original shape and valid shape.

        Returns:

        """
        coordinate_dims = [range(0, i, j) for i, j in zip(self.original_shape, self.valid_shape)]
        coordinates = np.meshgrid(*coordinate_dims, indexing="ij")
        coordinates = list(map(lambda x: x.flatten(), coordinates))
        # get the corner coordinate of each patch
        coordinates = np.stack(coordinates, axis=-1)
        coordinates = np.where(coordinates + self.valid_shape > self.original_shape,
                               self.original_shape - self.valid_shape, coordinates)

        # shift coordinate to patch center
        shift = get_margin(self.valid_shape)[:, 0]
        central_coordinates = coordinates + shift
        return central_coordinates.astype(np.int32)

    def __getitem__(self, index):
        return self.padded_coordinates[index]

    def __len__(self):
        return len(self.padded_coordinates)

    def regenerate(self):
        ...



from .brightness import MultiplicativeBrightnessTransform
from .contrast import ContrastTransform
from .data import ToType, ToTensor, AddChannel, Repeat
from .distance import SDF
from .downsample import DownsampleTransform
from .gamma import GammaTransform
from .gaussian_blur import GaussianBlurTransform
from .gaussian_noise import GaussianNoiseTransform
from .mirroring import MirrorTransform
from .normalization import ZscoreNormalization, MinMaxNormalization
from .spatial import SpatialTransform



import os
import re
import tomllib
from typing import Mapping


class Config(Mapping):
    def __init__(self, config: Mapping):
        self._config = config

    def __getattr__(self, name):
        try:
            value = self._config[name]
        except KeyError:
            raise AttributeError(f"The config has no key {name}") from None
        if isinstance(value, Mapping):
            return Config(value)
        return value

    def __getitem__(self, key):
        try:
            value = self._config[key]
        except KeyError:
            raise KeyError(f"The config has no key {key}") from None
        if isinstance(value, Mapping):
            return Config(value)
        return value

    def as_primitive(self) -> dict:
        return self._config

    def __dir__(self) -> list:
        return list(self._config.keys())

    def __repr__(self) -> str:
        return repr(self._config)

    def __str__(self) -> str:
        return f"configuration keys: {dir(self)}"

    def __len__(self):
        return len(self._config)

    def __iter__(self):
        return iter(self._config)


def load_config(config_file: str):
    if not os.path.isfile(config_file):
        raise FileNotFoundError(f"Config file {config_file} does not exist.")

    with open(config_file, "rb") as f:
        config = tomllib.load(f)
    refine_nodes(config, config)
    config = Config(config)
    return config


def refine_nodes(node, root):
    for k, v in node.items():
        if isinstance(v, str):
            replace_str(k, node, root)
        elif isinstance(v, dict):
            refine_nodes(v, root)
        elif isinstance(v, list):
            for i, inner_v in enumerate(v):
                if isinstance(inner_v, str):
                    replace_str(k, node, root, list_index=i)
                elif isinstance(inner_v, dict):
                    refine_nodes(inner_v, root)


def replace_str(key, node, root, list_index=None):
    """
    If `node[key]` is list type, set list_index to the index of current element of `node[key]`.

    Args:
        key (str):
        node (mapping):
        root (mapping):
        list_index (int or None, optional, default=None):

    Returns:

    """
    reference_patten = r"(\$\{.+?\})"
    reference_key_patten = r"\$\{(.+)\}"
    if list_index is not None:
        value = node[key][list_index]
    else:
        value = node[key]
    match = re.findall(reference_patten, value)
    if not match:
        return
    for each in match:
        replace_key = re.search(reference_key_patten, each).group(1)
        replace_node, replace_key = get_node_key(replace_key, node, root)
        if isinstance(replace_node[replace_key], str):
            replace_str(replace_key, replace_node, root)
        if list_index is not None:
            node[key][list_index] = node[key][list_index].replace(each, root[replace_key])
        else:
            node[key] = node[key].replace(each, str(replace_node[replace_key]))


def get_node_key(key, node, root):
    """
    Get the inner node and key.

    Args:
        key (str):
        node (dict):
        root (mapping):

    Returns:

    """
    node_hierarchy = key.split(".")
    # first, search on current node
    # second, search on root if search failed on current node

    node, unpacked_key = search_node(node_hierarchy, node)
    if node and unpacked_key:
        return node, unpacked_key

    node = root
    node, unpacked_key = search_node(node_hierarchy, node)
    if node and unpacked_key:
        return node, unpacked_key
    raise ValueError(f"The config has no key {key}")


def search_node(node_hierarchy, node):
    for i, key in enumerate(node_hierarchy):
        if key in node:
            if i == len(node_hierarchy) - 1:
                return node, node_hierarchy[-1]
            else:
                node = node[key]
    return None, None



import datetime
from enum import Enum
from random import randint

from pydicom import Dataset
from pydicom.dataset import FileDataset, FileMetaDataset
from pydicom.uid import generate_uid


class Modality(Enum):
    CT = 1


def get_dicom_dataset(name, modality: Modality) -> FileDataset:
    file_meta = FileMetaDataset()
    file_meta.TransferSyntaxUID = "1.2.840.10008.1.2.1"
    file_meta.ImplementationVersionName = "DICOM"

    ds = FileDataset(name, {}, file_meta=file_meta, preamble=b"\0" * 128)
    ds.is_implicit_VR = False
    ds.is_little_endian = True
    ds.ImageType = ["DERIVED", "SECONDARY"]

    # add properties
    patient(ds)
    general_study(ds)
    patient_study(ds)
    frame_of_reference(ds)
    general_equipment(ds)
    general_image(ds)
    general_acquisition(ds)
    image_plane(ds)
    image_pixel(ds)
    SOP_common(ds)
    VOI_LUT(ds)
    general_series(ds, file_meta, modality)

    # set datatime
    dt = datetime.datetime.now()
    date_str = dt.strftime("%Y%m%d")
    time_str = dt.strftime("%H%M%S.%f")  # long format with micro seconds

    ds.ContentDate = date_str
    ds.ContentTime = time_str
    ds.StudyDate = date_str
    ds.StudyTime = time_str
    ds.SeriesDate = date_str
    ds.SeriesTime = time_str
    ds.AcquisitionDate = date_str
    ds.AcquisitionTime = time_str
    ds.InstanceCreationDate = date_str
    ds.InstanceCreationTime = time_str

    ds.RescaleIntercept = ""
    ds.RescaleSlope = ""

    return ds


def patient(ds: Dataset):
    ds.PatientName = "Patient Name"
    ds.PatientID = "1"
    ds.PatientSex = ""
    ds.PatientBirthDate = ""


def general_study(ds: Dataset):
    ds.StudyInstanceUID = generate_uid()
    ds.StudyDescription = ""
    ds.ReferringPhysicianName = ""


def patient_study(ds: Dataset):
    ds.PatientAge = ""
    ds.PatientWeight = ""


def general_series(ds: Dataset, file_meta: FileMetaDataset, modality: Modality):
    ds.Modality = modality.name
    match modality:
        case Modality.CT:
            file_meta.MediaStorageSOPClassUID = "1.2.840.10008.5.1.4.1.1.2"
            ds.SOPClassUID = "1.2.840.10008.5.1.4.1.1.2"
            CT_image(ds)

    ds.SeriesInstanceUID = generate_uid(None)
    ds.SeriesNumber = str(randint(1000, 9999))
    ds.ProtocolName = "DICOM"
    ds.PatientPosition = ""
    ds.AccessionNumber = "123456"


def frame_of_reference(ds: Dataset):
    ds.FrameOfReferenceUID = generate_uid(None)


def general_equipment(ds: Dataset):
    ds.Manufacturer = ""
    ds.InstitutionName = "INSTITUTION_NAME_UNDEFINED"
    ds.ManufacturerModelName = ""
    ds.SoftwareVersions = ""


def general_image(ds: Dataset):
    ds.InstanceNumber = ""
    ds.PatientOrientation = ""
    ds.ContentDate = ""
    ds.ContentTime = ""
    ds.ImageType = ["SECONDARY", "DERIVED"]
    ds.LossyImageCompression = "00"


def general_acquisition(ds: Dataset):
    ds.AcquisitionNumber = ""
    ds.AcquisitionDate = ""
    ds.AcquisitionTime = ""


def image_plane(ds: Dataset):
    ds.PixelSpacing = ""
    ds.ImageOrientationPatient = ["1", "0", "0", "0", "1", "0"]
    ds.ImagePositionPatient = ["0", "0", "0"]
    ds.SliceThickness = ""
    ds.SpacingBetweenSlices = ""
    ds.SliceLocation = ""


def image_pixel(ds: Dataset):
    ds.Rows = 0
    ds.Columns = 0

    ds.BitsAllocated = 0
    ds.BitsStored = 0
    ds.HighBit = 0

    ds.PixelRepresentation = 1

    ds.SmallestImagePixelValue = ""
    ds.LargestImagePixelValue = ""

    ds.PixelData = ""


def SOP_common(ds: Dataset):
    ds.SOPClassUID = ""
    ds.SOPInstanceUID = ""

    ds.SpecificCharacterSet = "ISO_IR 100"
    ds.InstanceCreationDate = ""
    ds.InstanceCreationTime = ""

    ds.InstanceCreatorUID = ""


def VOI_LUT(ds: Dataset):
    ds.WindowCenter = ""
    ds.WindowWidth = ""


def CT_image(ds: Dataset):
    ds.SamplesPerPixel = 1

    ds.PhotometricInterpretation = "MONOCHROME2"

    ds.BitsAllocated = 16
    ds.BitsStored = 12
    ds.HighBit = ds.BitsStored - 1



import os.path
from typing import Union

import torch
from torch import nn
from torch.nn import DataParallel
from torch.nn.parallel import DistributedDataParallel
from torch.optim import Optimizer

from jbag.io import ensure_output_file_dir_existence
from jbag.log import logger

MODEL = "model"
OPTIMIZER = "optimizer"


def get_unwrapped_model(model: nn.Module):
    if isinstance(model, (DataParallel, DistributedDataParallel)):
        model = model.module
    return model


def save_weights(file: str, model: nn.Module, optimizer: Union[None, Optimizer] = None,
                 **kwargs):
    checkpoint = {MODEL: get_unwrapped_model(model).state_dict()}
    if optimizer:
        checkpoint[OPTIMIZER] = optimizer.state_dict()
    for k, v in kwargs.items():
        if k in checkpoint:
            raise KeyError(f"Get duplicated key {k}.")
        checkpoint[k] = v
    ensure_output_file_dir_existence(file)
    torch.save(checkpoint, file)


def load_weights(input_weights_file: str, model: Union[nn.Module, None] = None,
                 optimizer: Union[Optimizer, None] = None, map_location = None):
    if not os.path.isfile(input_weights_file):
        raise FileNotFoundError(f"Input weights file {input_weights_file} does not exist.")
    checkpoint = torch.load(input_weights_file, map_location=map_location)
    if model:
        if MODEL not in checkpoint:
            logger.warning(f"{input_weights_file} does not include model weights.")
        else:
            model = get_unwrapped_model(model)
            model.load_state_dict(checkpoint[MODEL])
            logger.info(f"Loading model weights from {input_weights_file}.")

    if optimizer:
        if OPTIMIZER not in checkpoint:
            logger.warning(f"{input_weights_file} does not include optimizer weights.")
        else:
            optimizer.load_state_dict(checkpoint[OPTIMIZER])
            logger.info(f"Loading optimizer weights from {input_weights_file}.")
    return checkpoint



import os
import shutil

import matplotlib.pyplot as plt
import nibabel as nib
import numpy as np
from pydicom import Dataset
from pydicom.uid import generate_uid

from jbag import logger
from jbag.converters.dicom import get_dicom_dataset, Modality


def write_slice(ds: Dataset, image_data, slice_index, output_dir):
    output_filename = r"Slice_%04d.dcm" % (slice_index + 1)
    image_slice = image_data[..., slice_index]
    ds.SOPInstanceUID = generate_uid(None)
    ds.PixelData = image_slice.tobytes()
    ds.save_as(os.path.join(output_dir, output_filename), write_like_original=False)


def get_nii2dcm_parameters(nii_data):
    nii_img = nii_data.get_fdata()

    dim_array = nii_data.header["dim"].astype(np.int16)
    num_x, num_y, num_z = int(dim_array[1]), int(dim_array[2]), int(dim_array[3])
    voxel_spacing_x, voxel_spacing_y, voxel_spacing_z = (nii_data.header["pixdim"][1],
                                                         nii_data.header["pixdim"][2],
                                                         nii_data.header["pixdim"][3])

    slice_indices = range(1, num_z + 1)
    last_location_z = (voxel_spacing_z * num_z) - voxel_spacing_z
    slice_locations = np.linspace(0, last_location_z, num=num_z)

    # Windowing
    max_i = np.max(nii_img)
    min_i = np.min(nii_img)
    window_center = round((max_i - min_i) / 2)
    window_width = round(max_i - min_i)

    rescale_intercept = 0
    rescale_slope = 1

    # FOV
    fov_x = num_x * voxel_spacing_x
    fov_y = num_y * voxel_spacing_y
    fov_z = num_z * voxel_spacing_z

    # slice positioning in 3-D space
    # -1 for direction cosines gives consistent orientation between Nifti and DICOM in ITK-Snap
    affine = nii_data.affine
    dir_cos_x = -1 * affine[:3, 0] / voxel_spacing_x
    dir_cos_y = -1 * affine[:3, 1] / voxel_spacing_y

    image_pos_patient_array = []
    for slice_i in range(0, num_z):
        v = affine.dot([0, 0, slice_i - 1, 1])
        image_pos_patient_array.append([v[0], v[1], v[2]])

    # output dictionary
    nii2dcm_parameters = {
        # series parameters
        "DimX": voxel_spacing_x,
        "DimY": voxel_spacing_y,
        "SliceThickness": str(voxel_spacing_z),
        "SpacingBetweenSlices": str(voxel_spacing_z),
        "AcquisitionMatrix": [0, num_x, num_y, 0],
        "Rows": num_x,
        "Columns": num_y,
        "NumberOfSlices": num_z,
        "NumberOfInstances": num_z,
        "PixelSpacing": [voxel_spacing_x, voxel_spacing_y],
        "FOV": [fov_x, fov_y, fov_z],
        "SmallestImagePixelValue": min_i,
        "LargestImagePixelValue": max_i,
        "WindowCenter": str(window_center),
        "WindowWidth": str(window_width),
        "RescaleIntercept": str(rescale_intercept),
        "RescaleSlope": str(rescale_slope),
        "ImageOrientationPatient": [dir_cos_y[0], dir_cos_y[1], dir_cos_y[2], dir_cos_x[0], dir_cos_x[1], dir_cos_x[2]],

        # instance parameters
        "InstanceNumber": slice_indices,
        "SliceLocation": slice_locations,
        "ImagePositionPatient": image_pos_patient_array
    }

    return nii2dcm_parameters


def transfer_dicom_series_tags(nii2dcm_parameters: dict, ds: Dataset):
    ds.Rows = nii2dcm_parameters["Rows"]
    ds.Columns = nii2dcm_parameters["Columns"]
    ds.PixelSpacing = [round(float(nii2dcm_parameters["DimX"]), 2), round(float(nii2dcm_parameters["DimY"]), 2)]
    ds.SliceThickness = nii2dcm_parameters["SliceThickness"]
    ds.SpacingBetweenSlices = round(float(nii2dcm_parameters["SpacingBetweenSlices"]), 2)
    ds.ImageOrientationPatient = nii2dcm_parameters["ImageOrientationPatient"]
    ds.AcquisitionMatrix = nii2dcm_parameters["AcquisitionMatrix"]
    ds.SmallestImagePixelValue = int(nii2dcm_parameters["SmallestImagePixelValue"]) \
        if int(nii2dcm_parameters["SmallestImagePixelValue"]) > 0 else 0
    ds.LargestImagePixelValue = int(nii2dcm_parameters["LargestImagePixelValue"])
    ds.WindowCenter = nii2dcm_parameters["WindowCenter"]
    ds.WindowWidth = nii2dcm_parameters["WindowWidth"]
    ds.RescaleIntercept = nii2dcm_parameters["RescaleIntercept"]
    ds.RescaleSlope = nii2dcm_parameters["RescaleSlope"]


def transfer_dicom_instance_tags(nii2dcm_parameters: dict, ds: Dataset, instance_index: int):
    ds.InstanceNumber = nii2dcm_parameters["InstanceNumber"][instance_index]
    ds.SliceLocation = nii2dcm_parameters["SliceLocation"][instance_index]
    ds.ImagePositionPatient = [
        str(nii2dcm_parameters["ImagePositionPatient"][instance_index][0]),
        str(nii2dcm_parameters["ImagePositionPatient"][instance_index][1]),
        str(nii2dcm_parameters["ImagePositionPatient"][instance_index][2]),
    ]


def nifti2dicom(input_nifti_file, output_dicom_dir, modality: Modality, force_overwrite=False):
    """
    Convert NIfTI image to DICOM image series. Inspired by https://github.com/tomaroberts/nii2dcm.

    Args:
        input_nifti_file (str):
        output_dicom_dir (str):
        modality (Modality):
        force_overwrite (bool, optional, default=False): If `Ture`, overwrite `output_dicom_dir` if it exists.`

    Returns:

    """

    if not os.path.isfile(input_nifti_file):
        raise FileNotFoundError(f"Input NIfTI file {input_nifti_file} does not exist.")

    if os.path.exists(output_dicom_dir):
        if force_overwrite:
            logger.info(f"Overwrite output DICOM dir {output_dicom_dir} as it already exists.")
            shutil.rmtree(output_dicom_dir)
        else:
            raise ValueError(f"Output DICOM series dir {output_dicom_dir} already exists.")

    nii_data = nib.load(input_nifti_file)

    nii2dcm_properties = get_nii2dcm_parameters(nii_data)
    file_name = os.path.splitext(os.path.basename(input_nifti_file))[0]

    dicom_ds = get_dicom_dataset(file_name, modality=modality)
    transfer_dicom_series_tags(nii2dcm_properties, dicom_ds)

    image = nii_data.get_fdata()
    image = image.astype(int)
    os.makedirs(output_dicom_dir)
    for instance_index in range(0, nii2dcm_properties["NumberOfInstances"]):
        transfer_dicom_instance_tags(nii2dcm_properties, dicom_ds, instance_index)
        write_slice(dicom_ds, image, instance_index, output_dicom_dir)



from typing import Union

from torch.nn.functional import interpolate

from jbag.transforms.transform import Transform


class DownsampleTransform(Transform):
    def __init__(self, keys, scales: Union[list, tuple]):
        """
        Down sample transform.
        Args:
            keys (str or sequence):
            scales (list, tuple): The element type should be float (used for all spatial dimensions) or list or tuple (each spatial dimension has a specific scale)
        """
        super().__init__(keys)
        self.scales = scales

    def _call_fun(self, data):
        for key in self.keys:
            value = data[key]
            results = []
            for scale in self.scales:
                if not isinstance(scale, (list, tuple)):
                    scale = [scale] * (value.ndim - 1)
                if len(scale) != value.ndim - 1:
                    raise ValueError(
                        f"Number of scales ({len(scale)}) must be equal to number of data dimensions ({value.ndim - 1}).")

                if all([i == 1 for i in scale]):
                    results.append(value)
                else:
                    new_shape = [round(i * j) for i, j in zip(value.shape[1:], scale)]
                    results.append(interpolate(value[None].float(), new_shape, mode="nearest-exact")[0].to(value.dtype))
            data[key] = results
        return data



import setuptools

setuptools.setup(
    name="jbag",
    version="4.2.7",
    author="Dai Jian",
    author_email="daijian@stumail.ysu.edu.cn",
    description="Tools for medical image processing and deap learning.",
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(exclude=["tests"]),
    classifiers=[
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires="~=3.11",
    install_requires=[
        "numpy >= 1.21.0",
        "nibabel >= 4.0.1",
        "scipy >= 1.7.3",
        "SimpleITK >= 2.1.1.2",
        "pydicom >= 2.4.3",
        "dicom2nifti >= 2.4.8",
        "pandas",
        "openpyxl",
        "python-docx",
        "scikit-image",
    ]
)



import numpy as np
import torch


class MetricSummary:
    def __init__(self, metric_fn=None):
        self.metric_fn = metric_fn
        self.__evaluations = []

    def __call__(self, input, target):
        evaluation = self.metric_fn(input, target)
        self.__evaluations.append(evaluation)
        return evaluation

    def reset(self):
        self.__evaluations = []

    def add_value(self, input_value):
        self.__evaluations.append(input_value)

    def mean(self):
        if self.__evaluations:
            if isinstance(self.__evaluations[0], torch.Tensor):
                return torch.stack(self.__evaluations).mean()
            else:
                return np.mean(self.__evaluations)
        else:
            raise ValueError("No evaluation exists.")

    def std(self):
        if self.__evaluations:
            if isinstance(self.__evaluations[0], torch.Tensor):
                return torch.stack(self.__evaluations).std()
            else:
                return np.std(self.__evaluations)
        else:
            raise ValueError("No evaluation exists.")

    def count(self):
        return len(self.__evaluations)

    @property
    def evaluations(self):
        return self.__evaluations



import torch
import torch.nn.functional as F


def gaussian_filter(input: torch.Tensor, sigma, axes=None, truncated=4):
    """
    Perform Gaussian filtering for the given array.
    This is a PyTorch implementation for image with data type of `torch.Tensor`.
    Args:
        input (torch.Tensor): Image to be filtered. The supported spatial dimensions of input is 1D, 2D, and 3D.
        An extra channel dimension should be attached before spatial dimensions in the input tensor, so the input tensor
        shape should be `[C, spatial_dims]`. This is because the data dimension requirement of PyTorch. Dimension `C`
        could be used for passing multiple tensor array.
        sigma (float or sequence): Standard deviation of the Gaussian kernel.
        If sigma is a sequence, then the number of elements must match those in axes sequence.
        axes (int or sequence, optional, default=None): axes for performing filtering.
        If None, Gaussian filtering will be performed on every axis.
        truncated (float, optional, default=6): Truncate the filter at this many standard deviations.

    Returns:
        Return tensor array of the same shape as input
    """
    if not (1 < input.dim() <= 4):
        raise ValueError(f"Dimensions of input must be in the range of (1, 4], got {input.dim()}.")

    if axes is None:
        axes = tuple(range(1, input.dim()))
    elif isinstance(axes, int):
        axes = [axes]

    for axis in axes:
        if not (0 < axis < input.dim()):
            raise ValueError(f"Axis must be in range of (0, {input.dim()}), got {axis}.")

    if isinstance(sigma, (list, tuple)):
        if len(sigma) != len(axes):
            raise ValueError(f"Number of sigmas must equal to the length of axes ({len(axes)}).")
    else:
        sigma = [sigma] * len(axes)

    # input.dim() - 1 is spatial dims
    conv_op = eval(f"F.conv{input.dim() - 1}d")

    for axis, s in zip(axes, sigma):
        kernel = _build_kernel(s, truncated, dtype=input.dtype)
        conv_weight_shape = [1] * (input.dim() + 1)
        conv_weight_shape[1 + axis] = len(kernel)
        conv_weight = kernel.view(*conv_weight_shape).expand(input.shape[0], *[-1] * input.dim())
        padding_shape = [0, 0] * (input.dim() - 1)
        padding_shape[2 * axis - 2:2 * axis] = [len(kernel) // 2] * 2
        padding_shape = padding_shape[::-1]
        input = F.pad(input, padding_shape, mode="reflect")
        input = conv_op(input=input, weight=conv_weight, groups=input.shape[0])
    return input


def _build_kernel(sigma: float, truncate: float, dtype):
    radius = int(sigma * truncate + 0.5)
    x = torch.arange(-radius, radius + 1, dtype=dtype)
    kernel = torch.exp(-0.5 * (x / sigma).pow(2))
    return kernel / kernel.sum()



from typing import Union

import torch

from jbag.image.gaussian_filter import gaussian_filter
from jbag.transforms._utils import get_scalar, get_max_spatial_dims
from jbag.transforms.transform import RandomTransform


class GaussianBlurTransform(RandomTransform):
    def __init__(self, keys,
                 apply_probability,
                 blur_sigma=Union[tuple[float], list[float]],
                 synchronize_channels: bool = False,
                 synchronize_axes: bool = False,
                 p_per_channel: float = 1):
        """
        Filter image using Gaussian filter.
        Args:
            keys (str or sequence):
            apply_probability (float):
            blur_sigma (sequence): Sigma for Gaussian blur. If sequence with two elements, Gaussian blur sigma is uniformly sampled from [blur_sigma[0], blur_sigma[1]).
            synchronize_channels (bool, optional, default=False): If True, use the same parameters for all channels.
            synchronize_axes (bool, optional, default=False): If True, use the same parameters for all axes of an image.
            p_per_channel (float, optional, default=1): Probability of applying transform to each channel.
        """
        super().__init__(keys, apply_probability)

        self.blur_sigma = blur_sigma
        self.synchronize_channels = synchronize_channels
        self.synchronize_axes = synchronize_axes
        self.p_per_channel = p_per_channel

    def _call_fun(self, data):
        apply_to_channel = torch.where(torch.rand(len(self.keys)) < self.p_per_channel)[0]
        if len(apply_to_channel) == 0:
            return data

        max_spatial_dims = get_max_spatial_dims(self.keys, apply_to_channel, data)

        if self.synchronize_axes:
            sigmas = [[get_scalar(self.blur_sigma)] * max_spatial_dims] * len(apply_to_channel) \
                if self.synchronize_channels else \
                [[get_scalar(self.blur_sigma)] * max_spatial_dims for _ in range(len(apply_to_channel))]
        else:
            sigmas = [[get_scalar(self.blur_sigma) for _ in range(max_spatial_dims)], ] * len(apply_to_channel) \
                if self.synchronize_channels else \
                [[get_scalar(self.blur_sigma) for _ in range(max_spatial_dims)] for _ in range(len(apply_to_channel))]

        for c, sigma in zip(apply_to_channel, sigmas):
            value = data[self.keys[c]]
            spatial_dim = value.shape[1:]
            sigma = sigma[:len(spatial_dim)]
            axes = list(range(1, len(spatial_dim) + 1))
            value = gaussian_filter(value, sigma=sigma, axes=axes)
            data[self.keys[c]] = value
        return data
